/*
 * ipvrf.c	"ip vrf"
 *
 *		This program is free software; you can redistribute it and/or
 *		modify it under the terms of the GNU General Public License
 *		as published by the Free Software Foundation; either version
 *		2 of the License, or (at your option) any later version.
 *
 * Authors:	David Ahern <dsa@cumulusnetworks.com>
 *
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/mount.h>
#include <linux/bpf.h>
#include <linux/if.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <limits.h>

#include "rt_names.h"
#include "utils.h"
#include "ip_common.h"

#define CGRP_PROC_FILE  "/cgroup.procs"
#define L3MDEV_FILE "/l3mdev.master-device"

static void usage(void)
{
	fprintf(stderr, "Usage: ip vrf exec [NAME] cmd ...\n");
	fprintf(stderr, "       ip vrf identify [PID]\n");
	fprintf(stderr, "       ip vrf pids [NAME]\n");

	exit(-1);
}

/* process /proc/mounts looking for path to l3mdev cgroup */
static char *find_l3mdev_cgroup(void)
{
	char path[4096];
	char fstype[128];    /* max length of any filesystem name */
	char *mnt = NULL;
	char *delim;
	FILE *fp;

	fp = fopen("/proc/mounts", "r");
	if (!fp) {
		fprintf(stderr,
			"Failed to open mounts file: %s\n", strerror(errno));
		return NULL;
	}

	while (fscanf(fp, "%*s %4096s %127s %*s %*d %*d\n",
		      path, fstype) == 2) {
		/* only care about cgroup filesystem */
		if (strcmp(fstype, "cgroup"))
			continue;

		/* check that path ends in "/l3mdev" */
		delim = strrchr(path, '/');
		if (!delim)
			continue;

		delim++;
		if (!strcmp(delim, "l3mdev")) {
			mnt = strdup(path);
			break;
		}
	}

	fclose(fp);

	if (!mnt) {
		fprintf(stderr, "Failed to find path to l3mdev cgroup. "
				"l3mdev controller enabled?\n");
	}

	return mnt;
}

/* return path to cgroups proc file for a vrf:
 * L3MDEV_MNT/VRF/cgroup.procs; for "default" vrf path
 * is L3MDEV_MNT/cgroup.procs
 */
static char *l3mdev_cgrp_procs(const char *vrf)
{
	size_t vrflen = 0, plen;
	char *mnt, *path = NULL;

	if (strcmp(vrf, "default"))
		vrflen = strlen(vrf) + 1; /* +1 for '/' */

	mnt = find_l3mdev_cgroup();
	if (!mnt)
		return NULL;

	plen = strlen(mnt) + vrflen + strlen(CGRP_PROC_FILE) + 1;
	path = malloc(plen);
	if (!path)
		goto out;

	strcpy(path, mnt);

	if (vrflen) {
		strcat(path, "/");
		strcat(path, vrf);
	}

	strcat(path, CGRP_PROC_FILE);

out:
	free(mnt);

	if (!path) {
		fprintf(stderr, "Failed to get cgroups file. "
				"l3mdev controller enabled?\n");
	}

	return path;
}

static int vrf_identify(pid_t pid, char *name, size_t len)
{
	char path[PATH_MAX];
	char buf[4096];
	char *vrf, *end;
	FILE *fp;

	snprintf(path, sizeof(path), "/proc/%d/cgroup", pid);
	fp = fopen(path, "r");
	if (!fp)
		return -1;

	memset(name, 0, len);

	while (fgets(buf, sizeof(buf), fp)) {
		vrf = strstr(buf, "l3mdev:/");
		if (vrf) {
			vrf += 8;  /* skip past l3mdev:/ */
			end = strchr(vrf, '\n');
			if (end)
				*end = '\0';

			strncpy(name, vrf, len - 1);
			break;
		}
	}

	fclose(fp);

	return 0;
}

static int ipvrf_identify(int argc, char **argv)
{
	unsigned int pid;
	char vrf[32];
	int rc;

	if (argc < 1)
		pid = getpid();
	else if (argc > 1)
		invarg("Extra arguments specified\n", argv[1]);
	else if (get_unsigned(&pid, argv[0], 10))
		invarg("Invalid pid\n", argv[0]);

	rc = vrf_identify(pid, vrf, sizeof(vrf));
	if (!rc) {
		if (vrf[0] != '\0')
			printf("%s\n", vrf);
	}

	return rc;
}

static int ipvrf_pids(int argc, char **argv)
{
	char buf[4096], *path;
	int fd, rc = -1;
	ssize_t n;

	if (argc != 1) {
		fprintf(stderr, "Invalid arguments\n");
		return -1;
	}

	path = l3mdev_cgrp_procs(argv[0]);
	if (!path)
		return -1;

	fd = open(path, O_RDONLY);
	if (fd < 0)
		goto out; /* no cgroup file, nothing to show */

	while (1) {
		n = read(fd, buf, sizeof(buf) - 1);
		if (n < 0) {
			fprintf(stderr,
				"Failed to read cgroups file: %s\n",
				strerror(errno));
			break;
		} else if (n == 0) {
			rc = 0;
			break;
		}
		printf("%s", buf);
	}

	close(fd);

out:
	free(path);
	return rc;
}

static int vrf_configure(const char *vrf)
{
	char path[PATH_MAX], *mnt;
	size_t vrflen;
	int ifindex;
	FILE *fp;

	/* nothing to do for default VRF */
	if (!strcmp(vrf, "default"))
		return 0;

	ifindex = name_is_vrf(vrf);
	if (!ifindex) {
		fprintf(stderr, "Invalid VRF name\n");
		return -1;
	}
	vrflen = strlen(vrf) + 1;

	/* lookup cgroup path */
	mnt = find_l3mdev_cgroup();
	if (!mnt)
		return -1;

	if (strlen(mnt) + vrflen + strlen(L3MDEV_FILE) + 1 > sizeof(path)) {
		fprintf(stderr, "Invalid mount path for cgroup\n");
		free(mnt);
		return -1;
	}
	snprintf(path, sizeof(path), "%s/%s", mnt, vrf);

	free(mnt);

	/* create directory for this vrf */
	if (mkdir(path, 0755) != 0) {
		if (errno == EEXIST)
			return 0;

		fprintf(stderr,
			"Failed to create vrf entry in cgroup: %s",
			strerror(errno));
		return -1;
	}

	strcat(path, L3MDEV_FILE);

	fp = fopen(path, "w+");
	if (!fp) {
		fprintf(stderr, "Failed to configure cgroup: %s\n",
			strerror(errno));
		return -1;
	}

	/* this cgroup wants the device name */
	fprintf(fp, "%s", vrf);
	fclose(fp);

	return 0;
}

static int ipvrf_configure(int argc, char **argv)
{
	if (argc != 1) {
		fprintf(stderr, "Invalid arguments\n");
		return -1;
	}

	return vrf_configure(argv[0]);
}

static int vrf_switch(const char *vrf)
{
	struct stat buf;
	int rc = -1;
	char *path;
	FILE *fp;

	if (strcmp(vrf, "default")) {
		if (!name_is_vrf(vrf)) {
			fprintf(stderr, "Invalid VRF name\n");
			return -1;
		}
	}

	path = l3mdev_cgrp_procs(vrf);
	if (!path)
		return -1;

	/* if cgroups.proc file does not exist then cgroup has
	 * not been configured for cgroup. Make it happen
	 */
	if (stat(path, &buf) != 0) {
		if (errno != ENOENT) {
			fprintf(stderr, "Failed to switch VRF: %s.\n",
				strerror(errno));
			goto out;
		}
		if (vrf_configure(vrf) != 0)
			goto out;
	}

	/*
	 * write pid to cgroup.procs making process part of cgroup
	 */
	fp = fopen(path, "w");
	if (!fp) {
		fprintf(stderr, "Failed to switch VRF: %s.\n", strerror(errno));
		goto out;
	}

	fprintf(fp, "%d", getpid());
	fclose(fp);

	rc = 0;
out:
	free(path);

	drop_cap();

	return rc;
}

static int ipvrf_exec(int argc, char **argv)
{
	if (argc < 1) {
		fprintf(stderr, "No VRF name specified\n");
		return -1;
	}
	if (argc < 2) {
		fprintf(stderr, "No command specified\n");
		return -1;
	}

	if (vrf_switch(argv[0]))
		return -1;

	return -cmd_exec(argv[1], argv + 1, !!batch_mode);
}

/* reset VRF association of current process to default VRF;
 * used by netns_exec
 */
void vrf_reset(void)
{
	char vrf[32];

	if (vrf_identify(getpid(), vrf, sizeof(vrf)) ||
	    (vrf[0] == '\0'))
		return;

	vrf_switch("default");
}

int do_ipvrf(int argc, char **argv)
{
	if (argc == 0) {
		fprintf(stderr, "No command given. Try \"ip vrf help\".\n");
		exit(-1);
	}

	if (matches(*argv, "identify") == 0)
		return ipvrf_identify(argc-1, argv+1);

	if (matches(*argv, "pids") == 0)
		return ipvrf_pids(argc-1, argv+1);

	if (matches(*argv, "exec") == 0)
		return ipvrf_exec(argc-1, argv+1);

	/* hidden command for ifupdown2 via vrf helper */
	if (matches(*argv, "configure") == 0)
		return ipvrf_configure(argc-1, argv+1);

	if (matches(*argv, "help") == 0)
		usage();

	fprintf(stderr, "Command \"%s\" is unknown, try \"ip vrf help\".\n",
		*argv);

	exit(-1);
}
